<!DOCTYPE html>
<html>
<head>
    <title>miniaudio - A single file audio playback and capture library.</title>
    <meta name="description" content="miniaudio is a single file audio playback and capture library written in C.">
	<meta name="viewport" content="width=device-width, initial-scale=1.0">
    
    <link rel="icon" href="{{ relative-path "img/favicon.png" }}">

{{ google-analytics }}

    <style>
{{ include "css/main.css" }}
    </style>
</head>

<body style="margin:0; padding:0">
    <div style="background-color:#003800; color:#bfa792;">
        <div style="max-width:100%; width:100%; margin:0 auto;">
            <table class="header-link-table" style="border-collapse:collapse; border-spacing:0; padding:0; padding-right:1em;">
                <tr>
                    <td style="padding:0.75em; width:100%; text-align:left;">
                        <table class="header-link-table" style="border-collapse:collapse; margin:0; padding:0">
                            <tr>
                                <td style="vertical-align:bottom; padding:0em; padding-right:2em;"><a href="{{ relative-path "index.html" }}"><img src="{{ relative-path "img/logo1_large_white.png" }}" style="height:24px; min-width:100%;"></a></td>
                                <td><a href="{{ relative-path "docs/manual/index.html" }}">Documentation</a></td>
                                <td><a href="{{ relative-path "docs/examples/index.html" }}">Examples</a></td>
                            </tr>
                        </table>    
                    </td>
                    
                    <td style="padding:0.1em; width:25%; text-align:right; vertical-align:center;">
                        <a href="https://www.reddit.com/r/miniaudio"><img src="{{ relative-path "img/reddit_white.svg" }}" style="margin:0; padding:0; height:40px; width:40px;"></a>
                    </td>
                    <td style="padding:0.1em; width:25%; text-align:right; vertical-align:center;">
                        <a href="https://discord.gg/9vpqbjU"><img src="{{ relative-path "img/Discord-Logo-White.svg" }}" style="margin:0; padding:0; height:32px; width:32px;"></a>
                    </td>
                    <td style="padding:0.1em; width:25%; text-align:right; vertical-align:center;">
                        <a rel="me" href="https://fosstodon.org/@mackron"><img src="{{ relative-path "img/mastodon_white.svg" }}" style="margin:0; padding:0; height:24px; width:32px;"></a>
                    </td>
                    <td style="padding:0.1em; padding-right:1em; width:25%; text-align:right; vertical-align:center;">
                        <a href="https://github.com/mackron/miniaudio"><img src="{{ relative-path "img/github_white.png" }}" style="margin:0; padding:0; height:24px; width:24px;"></a>
                    </td>
                </tr>
            </table>
        </div>
    </div>
    <div style="background-color:#fff; padding-bottom:0em; border-top:solid 1px #003800; background-color:#eee;">