{{ miniaudio-header }}
	<!-- Mobile redirection -->
	<script>
		//(function(a,b){if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino/i.test(a)||/1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substr(0,4)))window.location=b})(navigator.userAgent||navigator.vendor||window.opera,'mobile/index.html');
	</script>

    <div style="border-bottom:solid 1px #333; text-align:center; padding:64px; background-color:#eee;">
        <div><img src="img/logo1_large.png" style="width:auto; height:auto; min-height:70px;"></div>
        <div style="padding:16px; color:#333; font-size:14pt;">An audio playback and capture library for C and C++.</div>
		<div style="margin-top:32px;">
            <a class="a-download" href="https://raw.githubusercontent.com/mackron/miniaudio/master/miniaudio.h">Download miniaudio.h</a>
        </div>
    </div>
    
    <div style="margin:0 auto; text-align:center;">
		<div style="background-color:#fff; padding-top:3em; padding-bottom:4em; font-size:14pt; color:#666;">
			<div style="margin:0 auto; width:66%;">
                miniaudio is an audio playback and capture library for C and C++. It's made up of a single source
                file, has no external dependencies and is released into the public domain.
			</div>
			
			<table style="margin:0 auto; text-align:center; width:75%; margin-top:5em; font-size:12pt;">
				<tr>
					<td style="width:33%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Simple Build System</div>
						<div style="text-align:center;">
						Just add to your source tree and go. There's no need to install any dependencies or development
                        packages, nor are there any build systems to waste time on.
						</div>
					</td>
					<td style="width:34%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Cross Platform</div>
						<div style="text-align:center;">
						miniaudio works on all the major desktop and mobile platforms, including Windows, macOS,
                        Linux, BSD, iOS, Android and Web (via Emscripten).
						</div>
					</td>
					<td style="width:33%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Simple API</div>
						<div style="text-align:center;">
                        miniaudio has a simple, flexible and modular API that gets out of your way.
						</div>
					</td>
				</tr>
                
                <tr>
					<td style="width:33%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Open Source</div>
						<div style="text-align:center;">
                        miniaudio is open source with a permissive license of your choice of public domain or
                        <a href="https://github.com/aws/mit-0">MIT No Attribution.</a>
						</div>
					</td>
					<td style="width:34%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Detailed Documentation</div>
						<div style="text-align:center;">
						miniaudio has some of the best <a href="{{ relative-path "docs/manual/index.html" }}">documentation</a>
                        of any open source audio library and includes a suite of examples.
						</div>
					</td>
					<td style="width:33%; vertical-align:top; padding:2em;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">And Much More</div>
						<div style="text-align:center;">
                        Built-in decoders, advanced mixing and effect processing, resource management, 3D spatialization,
                        filters, data conversion and more.
						</div>
					</td>
				</tr>
			</table>
		</div>
        
        <div style="background-color:#eeeeee; border-top:solid 1px #333; padding-top:1em; padding-bottom:2em; font-size:14pt; color:#666;">
			<div class="feature-header" style="padding-top:1em;">Major Features</div>
			<table style="margin:0 auto; text-align:center; width:75%; margin-top:1em; font-size:12pt;">
				<tr>
					<td style="vertical-align:top; padding:3em; text-align:left">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Flexible and Modular API</div>
						<div style="padding-right:20em; text-align:justify; text-justify: inter-word;">
						miniaudio gives you complete flexibility. With the low level API, just initialize a connection
                        to the device and send or receive raw audio data. The modular design of miniaudio allows you to
                        use the low level API without compromising your ability to make use of other features like the
                        node graph and resource manager.
						</div>
					</td>
				</tr>
                
                <tr>
					<td style="vertical-align:top; padding:3em; text-align:right;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em; text-alight:right;">Advanced Mixing and Effect Processing</div>
						<div style="padding-left:20em; text-align:justify; text-justify: inter-word;">
                        miniaudio's node graph system gives you an easy way to set up advanced mixing and effect
                        graphs. Each node sends it's output to another node, and so on and so forth to produce all
                        kinds of effects. You can even implement your own custom nodes and plug them in however you
                        like. Plug multiple nodes into the same input node for mixing and use the splitter node for
                        complex routing.
						</div>
					</td>
				</tr>
                
                <tr>
					<td style="vertical-align:top; padding:3em; text-align:left">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em;">Resource Management</div>
						<div style="padding-right:20em; text-align:justify; text-justify: inter-word;">
						The resource manager is used for simplifying the hassle of dealing with the loading and management
                        of your audio resources. It will reference count files so they're only loaded once and handles
                        streaming of large audio sources to save on memory. It can even load files asynchronously and
                        exposes its job system so you can handle resource management jobs from your existing job
                        infrastructure.
						</div>
					</td>
				</tr>
                
                <tr>
					<td style="vertical-align:top; padding:3em; text-align:right;">
						<div style="color:#464; font-size:18pt; font-weight:bold; margin-bottom:1em; text-alight:right;">High Level Audio Engine</div>
						<div style="padding-left:20em; text-align:justify; text-justify: inter-word;">
                        The high level audio engine in miniaudio encapsulates the resource manager and node graph to give
                        you an easy to use API to manage sounds within a 3D scene with spatialization. The engine is a
                        node graph, and each sound in the scene is a node within that graph which means you can take
                        advantage of miniaudio's graph based effect processing and routing infrastructure.
						</div>
					</td>
				</tr>
			</table>
		</div>
	
        <div style="background-color:#ffffff; border-top:solid 1px #333; padding-top:1em; padding-bottom:2em; font-size:14pt; color:#666;">
			<div class="feature-header" style="padding-top:1em;">Platforms</div>
			<table style="margin:0 auto; text-align:center; width:75%; margin-top:2em; font-size:12pt;">
				<tr>
					<td style="width:20%; vertical-align:center; padding:1em;">
						<img src="{{ relative-path "img/platforms/windows/windows-10.svg" }}" style="height:128px;">
					</td>
                    <td style="width:20%; vertical-align:center; padding:1em;">
						<img src="{{ relative-path "img/platforms/apple/apple-black.svg" }}" style="height:128px;">
					</td>
                    <td style="width:20%; vertical-align:center; padding:1em;">
						<img src="{{ relative-path "img/platforms/linux/linux.svg" }}" style="height:128px;">
					</td>
                    <td style="width:20%; vertical-align:center; padding:1em;">
						<img src="{{ relative-path "img/platforms/bsd/freebsd.svg" }}" style="height:128px;">
					</td>
                    <td style="width:20%; vertical-align:center; padding:1em;">
                        <img src="{{ relative-path "img/platforms/android/android.svg" }}" style="height:100px;">
					</td>
				</tr>
                
                <tr style="font-weight:bold;">
					<td style="vertical-align:top; padding:0em; font-weight:bold;">
						Windows
					</td>
                    <td style="vertical-align:top; padding:0em;">
						macOS / iOS
					</td>
                    <td style=vertical-align:top; padding:0em;">
						Linux
					</td>
                    <td style="vertical-align:top; padding:0em;">
						FreeBSD / 
                        OpenBSD /
                        NetBSD
					</td>
                    <td style="vertical-align:top; padding:0em;">
                        Android
					</td>
				</tr>
                
                <tr>
					<td style="vertical-align:top; padding-top:1em;">
                        WASAPI<br/>
                        DirectSound<br/>
                        WinMM
					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                        Core Audio
					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                        ALSA<br/>
                        PulseAudio<br/>
                        JACK
					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                        OSS<br/>
                        sndio<br/>
                        audio(4)
					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                        AAudio<br/>
                        OpenSL | ES
					</td>
				</tr>
                
                
                <!-- PLATFORM ROW 2 -->
                <tr>
					<td style="width:20%; vertical-align:center; padding:1em; padding-top:2em;">

					</td>
                    <td style="width:20%; vertical-align:center; padding:1em; padding-top:2em;">

					</td>
                    <td style="width:20%; vertical-align:center; padding:1em; padding-top:4em;">
						<img src="{{ relative-path "img/platforms/web/html5.svg" }}" style="height:128px;">
					</td>
                    <td style="width:20%; vertical-align:center; padding:1em; padding-top:2em;">

					</td>
                    <td style="width:20%; vertical-align:center; padding:1em; padding-top:2em;">

					</td>
				</tr>
                
                <tr style="font-weight:bold;">
					<td style="vertical-align:top; padding:0em; font-weight:bold;">

					</td>
                    <td style="vertical-align:top; padding:0em;">

					</td>
                    <td style=vertical-align:top; padding:0em;">
						Web
					</td>
                    <td style="vertical-align:top; padding:0em;">

					</td>
                    <td style="vertical-align:top; padding:0em;">

					</td>
				</tr>
                
                <tr>
					<td style="vertical-align:top; padding-top:1em;">

					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                    
					</td>
                    <td style="vertical-align:top; padding-top:1em;">
                        Emscripten / WebAudio
					</td>
                    <td style="vertical-align:top; padding-top:1em;">

					</td>
                    <td style="vertical-align:top; padding-top:1em;">

					</td>
				</tr>
			</table>
		</div>
    </div>
{{ miniaudio-footer }}
